@@ -20,17 +20,18 @@
 
 import apt_pkg
 import re
+import fnmatch
 import logging
 
-PROGRESS = 15
-
-logging.addLevelName(logging.DEBUG, '  ')
-logging.addLevelName(PROGRESS, '')
-logging.addLevelName(logging.INFO, '* ')
-logging.addLevelName(logging.WARNING, '! ')
-logging.addLevelName(logging.ERROR, '? ')
+try:
+    set # introduced in 2.4
+except NameError:
+    import sets
+    set = sets.Set
 
 class Germinator:
+    PROGRESS = 15
+
     def __init__(self):
         self.packages = {}
         self.packagetype = {}
@@ -40,6 +41,7 @@ class Germinator:
 
         self.seeds = []
         self.seed = {}
+        self.seedrecommends = {}
         self.seedinherit = {}
         self.seedrelease = {}
         self.substvars = {}
@@ -51,11 +53,11 @@ class Germinator:
 
         self.pkgprovides = {}
 
-        self.all = []
+        self.all = set()
         self.build = {}
         self.not_build = {}
 
-        self.all_srcs = []
+        self.all_srcs = set()
         self.build_srcs = {}
         self.not_build_srcs = {}
 
@@ -65,15 +67,17 @@ class Germinator:
         self.hints = {}
 
         self.blacklist = {}
-        self.blacklisted = []
+        self.blacklisted = set()
 
         self.di_kernel_versions = {}
+        self.extra_includes = {}
+        self.extra_excludes = {}
 
     def debug(self, msg, *args, **kwargs):
         logging.debug(msg, *args, **kwargs)
 
     def progress(self, msg, *args, **kwargs):
-        logging.getLogger().log(PROGRESS, msg, *args, **kwargs)
+        logging.getLogger().log(self.PROGRESS, msg, *args, **kwargs)
 
     def info(self, msg, *args, **kwargs):
         logging.info(msg, *args, **kwargs)
@@ -133,7 +137,7 @@ class Germinator:
             pkg = p.Section["Package"]
             self.packages[pkg] = {}
             self.packagetype[pkg] = pkgtype
-            self.pruned[pkg] = []
+            self.pruned[pkg] = set()
 
             self.packages[pkg]["Maintainer"] = \
                 unicode(p.Section.get("Maintainer", ""), "utf8", "replace")
@@ -214,7 +218,9 @@ class Germinator:
         """Write out the list of blacklisted packages we encountered"""
 
         fh = open(filename, 'w')
-        for pkg in self.blacklisted:
+        sorted_blacklisted = list(self.blacklisted)
+        sorted_blacklisted.sort()
+        for pkg in sorted_blacklisted:
             blacklist = self.blacklist[pkg]
             fh.write('%s\t%s\n' % (pkg, blacklist))
         fh.close()
@@ -222,18 +228,40 @@ class Germinator:
     def newSeed(self, seedname, seedinherit, seedrelease):
         self.seeds.append(seedname)
         self.seed[seedname] = []
+        self.seedrecommends[seedname] = []
         self.seedinherit[seedname] = seedinherit
         self.seedrelease[seedname] = seedrelease
-        self.depends[seedname] = []
-        self.build_depends[seedname] = []
-        self.sourcepkgs[seedname] = []
-        self.build_sourcepkgs[seedname] = []
-        self.build[seedname] = []
-        self.not_build[seedname] = []
-        self.build_srcs[seedname] = []
-        self.not_build_srcs[seedname] = []
+        self.depends[seedname] = set()
+        self.build_depends[seedname] = set()
+        self.sourcepkgs[seedname] = set()
+        self.build_sourcepkgs[seedname] = set()
+        self.build[seedname] = set()
+        self.not_build[seedname] = set()
+        self.build_srcs[seedname] = set()
+        self.not_build_srcs[seedname] = set()
         self.why[seedname] = {}
-        self.di_kernel_versions[seedname] = []
+        self.di_kernel_versions[seedname] = set()
+        self.extra_includes[seedname] = []
+        self.extra_excludes[seedname] = []
+
+    def filterPackages(self, packages, pattern):
+        """Filter a list of packages, returning those that match the given
+        pattern. The pattern may either be a shell-style glob, or (if
+        surrounded by slashes) an extended regular expression."""
+
+        if pattern.startswith('/') and pattern.endswith('/'):
+            patternre = re.compile(pattern[1:-1])
+            filtered = [p for p in packages if patternre.search(p) is not None]
+        elif '*' in pattern or '?' in pattern or '[' in pattern:
+            filtered = fnmatch.filter(packages, pattern)
+        else:
+            # optimisation for common case
+            if pattern in packages:
+                filtered = [pattern]
+            else:
+                filtered = []
+        filtered.sort()
+        return filtered
 
     def substituteSeedVars(self, pkg):
         """Process substitution variables. These look like ${name} (e.g.
@@ -291,7 +319,8 @@ class Germinator:
         which we inherit?"""
 
         for seed in self.innerSeeds(seedname):
-            if pkg in self.seed[seed]:
+            if (pkg in self.seed[seed] or
+                pkg in self.seedrecommends[seed]):
                 return True
 
         return False
@@ -303,6 +332,7 @@ class Germinator:
 
         self.newSeed(seedname, seedinherit, seedrelease)
         seedpkgs = []
+        seedrecommends = []
 
         for line in entries:
             if not line.startswith(" * "):
@@ -322,7 +352,13 @@ class Germinator:
                 if name == "kernel-version":
                     # Allows us to pick the right modules later
                     self.warning("Allowing d-i kernel versions: %s", values)
-                    self.di_kernel_versions[seedname].extend(values)
+                    self.di_kernel_versions[seedname].update(values)
+                elif name == "extra-include":
+                    self.warning("Including packages from extra: %s", values)
+                    self.extra_includes[seedname].extend(values)
+                elif name == "extra-exclude":
+                    self.warning("Excluding packages from extra: %s", values)
+                    self.extra_excludes[seedname].extend(values)
                 self.substvars[name] = values
                 continue
 
@@ -339,19 +375,28 @@ class Germinator:
 
             pkg = pkg.split()[0]
 
+            # a (pkgname) indicates that this is a recommend
+            # and not a depends
+            if pkg.startswith('(') and pkg.endswith(')'):
+                pkg = pkg[1:-1]
+                pkgs =  self.filterPackages(self.packages, pkg)
+                if not pkgs:
+                    pkgs = [pkg] # virtual or expanded; check again later
+                for pkg in pkgs:
+                    seedrecommends.extend(self.substituteSeedVars(pkg))
+
             if pkg.startswith('%'):
                 pkg = pkg[1:]
                 if pkg in self.sources:
-                    pkgs = list(self.sources[pkg]["Binaries"])
+                    pkgs = [p for p in self.sources[pkg]["Binaries"]
+                              if p in self.packages]
                 else:
                     self.warning("Unknown source package: %s", pkg)
                     pkgs = []
-            elif pkg.startswith('/') and pkg.endswith('/'):
-                pkgre = re.compile(pkg[1:-1])
-                pkgs = [p for p in self.packages if pkgre.search(p) is not None]
-                pkgs.sort()
             else:
-                pkgs = [pkg]
+                pkgs = self.filterPackages(self.packages, pkg)
+                if not pkgs:
+                    pkgs = [pkg] # virtual or expanded; check again later
 
             for pkg in pkgs:
                 seedpkgs.extend(self.substituteSeedVars(pkg))
@@ -365,22 +410,28 @@ class Germinator:
                 # Ordinary package
                 if self.alreadySeeded(seedname, pkg):
                     self.warning("Duplicated seed: %s", pkg)
-                elif seedname in self.pruned[pkg]:
+                elif self.is_pruned(pkg, seedname):
                     self.warning("Pruned %s from %s", pkg, seedname)
                 else:
-                    self.seed[seedname].append(pkg)
-
+                    if pkg in seedrecommends:
+                        self.seedrecommends[seedname].append(pkg)
+                    else:
+                        self.seed[seedname].append(pkg)
             elif pkg in self.provides:
                 # Virtual package, include everything
-                self.info("Virtual %s package: %s", seedname, pkg)
+                msg = "Virtual %s package: %s" % (seedname, pkg)
                 for vpkg in self.provides[pkg]:
                     if self.alreadySeeded(seedname, vpkg):
                         pass
                     elif seedname in self.pruned[vpkg]:
                         pass
                     else:
-                        self.debug("- %s", vpkg)
-                        self.seed[seedname].append(vpkg)
+                        msg += "\n  - %s" % vpkg
+                        if pkg in seedrecommends:
+                            self.seedrecommends[seedname].append(vpkg)
+                        else:
+                            self.seed[seedname].append(vpkg)
+                self.info("%s", msg)
 
             else:
                 # No idea
@@ -389,53 +440,70 @@ class Germinator:
         for pkg in self.hints:
             if self.hints[pkg] == seedname and not self.alreadySeeded(seedname, pkg):
                 if pkg in self.packages:
-                    self.seed[seedname].append(pkg)
+                    if pkg in seedrecommends:
+                        self.seedrecommends[seedname].append(pkg)
+                    else:
+                        self.seed[seedname].append(pkg)
                 else:
                     self.error("Unknown hinted package: %s", pkg)
 
+    def is_pruned(self, pkg, seed):
+        if not self.di_kernel_versions[seed]:
+            return False
+        kernver = self.packages[pkg]["Kernel-Version"]
+        if kernver != "" and kernver not in self.di_kernel_versions[seed]:
+            return True
+        return False
+
     def prune(self):
         """Remove packages that are inapplicable for some reason, such as
            being for the wrong d-i kernel version."""
         for pkg in self.packages:
-            kernver = self.packages[pkg]["Kernel-Version"]
-            if kernver != "":
-                for seed in self.seeds:
-                    if kernver not in self.di_kernel_versions[seed]:
-                        self.pruned[pkg].append(seed)
+            for seed in self.seeds:
+                if self.is_pruned(pkg, seed):
+                    self.pruned[pkg].add(seed)
 
     def grow(self):
         """Grow the seeds."""
         for seedname in self.seeds:
             self.progress("Resolving %s dependencies ...", seedname)
-            for pkg in self.seed[seedname]:
+            for pkg in self.seed[seedname] + self.seedrecommends[seedname]:
                 if self.seedrelease[seedname] is None:
                     self.addPackage(seedname, pkg, "%s seed" %
                         seedname.title())
                 else:
                     self.addPackage(seedname, pkg, "%s %s seed" %
                         (self.seedrelease[seedname].title(), seedname))
+            self.rescueExtras(seedname, build_tree=False)
+        self.rescueExtras("supported", build_tree=True)
 
     def addExtras(self, seedrelease=None):
         """Add packages generated by the sources but not in any seed."""
         self.newSeed("extra", self.seeds, seedrelease)
 
         self.progress("Identifying extras ...")
-        for srcname in self.all_srcs:
-            for pkg in self.sources[srcname]["Binaries"]:
-                if pkg not in self.packages:
-                    continue
-                if self.packages[pkg]["Source"] != srcname:
-                    continue
-                if pkg in self.all:
-                    continue
+        found = True
+        while found:
+            found = False
+            sorted_srcs = list(self.all_srcs)
+            sorted_srcs.sort()
+            for srcname in sorted_srcs:
+                for pkg in self.sources[srcname]["Binaries"]:
+                    if pkg not in self.packages:
+                        continue
+                    if self.packages[pkg]["Source"] != srcname:
+                        continue
+                    if pkg in self.all:
+                        continue
 
-                if pkg in self.hints and self.hints[pkg] != "extra":
-                    self.warning("Taking the hint: %s", pkg)
-                    continue
+                    if pkg in self.hints and self.hints[pkg] != "extra":
+                        self.warning("Taking the hint: %s", pkg)
+                        continue
 
-                self.seed["extra"].append(pkg)
-                self.addPackage("extra", pkg, "Generated by " + srcname,
-                                second_class=True)
+                    self.seed["extra"].append(pkg)
+                    self.addPackage("extra", pkg, "Generated by " + srcname,
+                                    second_class=True)
+                    found = True
 
     def allowedDependency(self, pkg, depend, seedname, build_depend):
         """Is pkg allowed to satisfy a (build-)dependency using depend
@@ -529,7 +597,8 @@ class Germinator:
                 for seed in self.innerSeeds(seedname):
                     if trydep in self.not_build[seed]:
                         return True
-            if trydep in self.seed[seedname]:
+            if (trydep in self.seed[seedname] or
+                trydep in self.seedrecommends[seedname]):
                 return True
         else:
             return False
@@ -556,7 +625,8 @@ class Germinator:
         found = False
         for trydep in trylist:
             for lesserseed in self.strictlyOuterSeeds(seedname):
-                if trydep in self.seed[lesserseed]:
+                if (trydep in self.seed[lesserseed] or
+                    trydep in self.seedrecommends[lesserseed]):
                     if second_class:
                         # "I'll get you next time, Gadget!"
                         # When processing the build tree, we don't promote
@@ -568,7 +638,10 @@ class Germinator:
                         # an or-ed build-dependency.
                         pass
                     else:
-                        self.seed[lesserseed].remove(trydep)
+                        if trydep in self.seed[lesserseed]:
+                            self.seed[lesserseed].remove(trydep)
+                        if trydep in self.seedrecommends[lesserseed]:
+                            self.seedrecommends[lesserseed].remove(trydep)
                         self.warning("Promoted %s from %s to %s to satisfy %s",
                                      trydep, lesserseed, seedname, pkg)
 
@@ -587,7 +660,8 @@ class Germinator:
                 # for other allowed kernel versions too.
                 if self.packages[depname]["Kernel-Version"] != "":
                     dependlist = [ d for d in reallist
-                                   if self.packages[d]["Kernel-Version"] in self.di_kernel_versions[seedname] ]
+                                   if not self.di_kernel_versions[seedname] or
+                                      self.packages[d]["Kernel-Version"] in self.di_kernel_versions[seedname] ]
                 else:
                     dependlist = [depname]
                 self.info("Chose %s out of %s to satisfy %s",
@@ -599,14 +673,14 @@ class Germinator:
 
         if build_tree:
             for dep in dependlist:
-                self.build_depends[seedname].append(dep)
+                self.build_depends[seedname].add(dep)
             if build_depend:
                 why = self.packages[pkg]["Source"] + " (Build-Depend)"
             else:
                 why = pkg
         else:
             for dep in dependlist:
-                self.depends[seedname].append(dep)
+                self.depends[seedname].add(dep)
             why = pkg
 
         for dep in dependlist:
@@ -660,24 +734,23 @@ class Germinator:
         if build_tree: second_class=True
 
         if pkg not in self.all:
-            self.all.append(pkg)
+            self.all.add(pkg)
         elif not build_tree:
             for buildseed in self.innerSeeds(seedname):
-                if pkg in self.build_depends[buildseed]:
-                    self.build_depends[buildseed].remove(pkg)
+                self.build_depends[buildseed].discard(pkg)
 
         for seed in self.innerSeeds(seedname):
             if pkg in self.build[seed]:
                 break
         else:
-            self.build[seedname].append(pkg)
+            self.build[seedname].add(pkg)
 
         if not build_tree:
             for seed in self.innerSeeds(seedname):
                 if pkg in self.not_build[seed]:
                     break
             else:
-                self.not_build[seedname].append(pkg)
+                self.not_build[seedname].add(pkg)
 
         # Remember why the package was added to the output for this seed.
         # Also remember a reason for "all" too, so that an aggregated list
@@ -687,9 +760,8 @@ class Germinator:
 
         for prov in self.packages[pkg]["Provides"]:
             if prov[0][0] not in self.pkgprovides:
-                self.pkgprovides[prov[0][0]] = []
-            if pkg not in self.pkgprovides[prov[0][0]]:
-                self.pkgprovides[prov[0][0]].append(pkg)
+                self.pkgprovides[prov[0][0]] = set()
+            self.pkgprovides[prov[0][0]].add(pkg)
 
         self.addDependencyTree(seedname, pkg,
                                self.packages[pkg]["Pre-Depends"],
@@ -715,22 +787,20 @@ class Germinator:
                     return
 
         if build_tree:
-            self.build_sourcepkgs[seedname].append(src)
-            if src in self.blacklist and src not in self.blacklisted:
-                self.blacklisted.append(src)
+            self.build_sourcepkgs[seedname].add(src)
+            if src in self.blacklist:
+                self.blacklisted.add(src)
 
         else:
             if src in self.all_srcs:
                 for buildseed in self.seeds:
-                    if src in self.build_sourcepkgs[buildseed]:
-                        self.build_sourcepkgs[buildseed].remove(src)
+                    self.build_sourcepkgs[buildseed].discard(src)
 
-            self.not_build_srcs[seedname].append(src)
-            self.sourcepkgs[seedname].append(src)
+            self.not_build_srcs[seedname].add(src)
+            self.sourcepkgs[seedname].add(src)
 
-        if src not in self.all_srcs:
-            self.all_srcs.append(src)
-        self.build_srcs[seedname].append(src)
+        self.all_srcs.add(src)
+        self.build_srcs[seedname].add(src)
 
         self.addDependencyTree(seedname, pkg,
                                self.sources[src]["Build-Depends"],
@@ -738,3 +808,49 @@ class Germinator:
         self.addDependencyTree(seedname, pkg,
                                self.sources[src]["Build-Depends-Indep"],
                                build_depend=True)
+
+    def rescueExtras(self, seedname, build_tree):
+        """Automatically rescue packages matching certain patterns from
+        extra."""
+
+        # Find all the source packages.
+        extra_srcs = set()
+        for seed in self.innerSeeds(seedname):
+            if build_tree:
+                extra_srcs |= self.build_srcs[seed]
+            else:
+                extra_srcs |= self.not_build_srcs[seed]
+
+        # For each source, add any binaries that match the include/exclude
+        # patterns.
+        for src in extra_srcs:
+            extras = [p for p in self.sources[src]["Binaries"]
+                        if p in self.packages]
+            included_extras = set()
+            for include in self.extra_includes[seedname]:
+                included_extras |= set(self.filterPackages(extras, include))
+            for exclude in self.extra_excludes[seedname]:
+                included_extras -= set(self.filterPackages(extras, exclude))
+            for pkg in included_extras:
+                if pkg in self.all:
+                    continue
+                for lesserseed in self.strictlyOuterSeeds(seedname):
+                    if pkg in self.seed[lesserseed]:
+                        self.seed[lesserseed].remove(pkg)
+                        self.warning("Promoted %s from %s to %s due to "
+                                     "Extra-Includes",
+                                     pkg, lesserseed, seedname)
+                        break
+                self.debug("Rescued %s from extra to %s", pkg, seedname)
+                if build_tree:
+                    self.build_depends[seedname].add(pkg)
+                else:
+                    self.depends[seedname].add(pkg)
+                self.addPackage(seedname, pkg, "Rescued from %s" % src,
+                                build_tree=build_tree)
+
+logging.addLevelName(logging.DEBUG, '  ')
+logging.addLevelName(Germinator.PROGRESS, '')
+logging.addLevelName(logging.INFO, '* ')
+logging.addLevelName(logging.WARNING, '! ')
+logging.addLevelName(logging.ERROR, '? ')
