@@ -24,30 +24,18 @@
 import os, sys, getopt
 import logging
 import apt_pkg
-import urllib
 from Germinate import Germinator
 import Germinate.Archive
+import Germinate.seeds
 
 # TODO: cloned from germinate.py; should be common
 SEEDS = "http://people.ubuntu.com/~cjwatson/seeds/"
-RELEASE = "dapper"
+RELEASE = "edgy"
 MIRROR = "http://archive.ubuntu.com/ubuntu/"
-DIST = ["dapper"]
+DIST = ["edgy"]
 COMPONENTS = ["main"]
 ARCH = "i386"
 
-# TODO: cloned from germinate.py; should be common
-def open_metafile(filename):
-    try:
-        url = SEEDS + RELEASE + "/" + filename
-        print "Downloading", url, "..."
-        urlopener = urllib.FancyURLopener()
-        urlopener.addheader('Cache-Control', 'no-cache')
-        urlopener.addheader('Pragma', 'no-cache')
-        return urlopener.open(url)
-    except IOError:
-        return None
-
 class Package:
     def __init__(self, name):
         self.name = name
@@ -110,10 +98,11 @@ class Globals:
 
         Germinate.Archive.TagFile(MIRROR).feed(g, DIST, COMPONENTS, ARCH)
 
-        (seednames, seedinherit) = g.parseStructure(open_metafile("STRUCTURE"))
+        (seednames, seedinherit) = g.parseStructure(
+            Germinate.seeds.open_seed(SEEDS + RELEASE, "STRUCTURE"))
         for seedname in self.seeds:
-            g.plantSeed(open_metafile(seedname), ARCH, seedname,
-                        list(seedinherit[seedname]), RELEASE)
+            g.plantSeed(Germinate.seeds.open_seed(SEEDS + RELEASE, seedname),
+                        ARCH, seedname, list(seedinherit[seedname]), RELEASE)
         g.prune()
         g.grow()
 
